"""
 Modified version of Original code https://github.com/ischlag/tensorflow-input-pipelines/blob/master/utils/download.py

"""
import sys
import os
import urllib.request
import tarfile
import numpy as np


def _print_download_progress(count, block_size, total_size):
    """
    Function used for printing the download progress.
    Used as a call-back function in maybe_download_and_extract().
    """

    # Percentage completion.
    pct_complete = float(count * block_size) / total_size

    # Status-message. Note the \r which means the line should overwrite itself.
    msg = "\r- Download progress: {0:.1%}".format(pct_complete)

    # Print it.
    sys.stdout.write(msg)
    sys.stdout.flush()


def maybe_download_and_extract(url, download_dir):
    """Download and extract the data if it doesn't already exist.
    Assumes the url is a tar-ball file.

    Args:
        url: Internet URL for the tar-file to download.
            Example: "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
        download_dir: Directory where the downloaded file is saved.
            Example: "data/CIFAR-10/"
    Return:
        Nothing.
    """

    # Filename for saving the file downloaded from the internet.
    # Use the filename from the URL and add it to the download_dir.
    filename = url.split('/')[-1]
    file_path = os.path.join(download_dir, filename)

    # Check if the download directory exists, otherwise create it.
    # If it exists then we assume it has also been extracted,
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

        # Download the file from the internet.
        file_path, _ = urllib.request.urlretrieve(url=url,
                                                  filename=file_path,
                                                  reporthook=_print_download_progress)

        print()
        print("Download finished. Extracting files.")

        # Unpack the tar-ball.
        tarfile.open(name=file_path, mode="r:gz").extractall(download_dir)
        print("Extracting finished. Cleaning up.")
        os.remove(file_path)
        print("Done.")
        return True
    else:
        print("Data has apparently already been downloaded and unpacked.")
        return False


def maybe_download(url, download_dir):
    """Download and extract the data if it doesn't already exist.
    Assumes the url is a tar-ball file.

    Args:
        url: Internet URL for the tar-file to download.
            Example: "https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz"
        download_dir: Directory where the downloaded file is saved.
            Example: "data/CIFAR-10/"
    Return:
        Nothing.
    """

    # Filename for saving the file downloaded from the internet.
    # Use the filename from the URL and add it to the download_dir.
    filename = url.split('/')[-1]
    file_path = os.path.join(download_dir, filename)

    # Check if the file already exists.
    # If it exists then we assume it has also been extracted,
    # otherwise we need to download and extract it now.
    if not os.path.exists(file_path):
        # Check if the download directory exists, otherwise create it.
        if not os.path.exists(download_dir):
            os.makedirs(download_dir)

        # Download the file from the internet.
        file_path, _ = urllib.request.urlretrieve(url=url,
                                                  filename=file_path,
                                                  reporthook=_print_download_progress)

        print()
        print("Download finished.")
        print("Done.")
        return True
    else:
        print("Data has apparently already been downloaded and unpacked.")
        return False


def one_hot_encoded(class_numbers, num_classes=None):
    """Generate the One-Hot encoded class-labels from an array of integers.
    For example, if class_number=2 and num_classes=4 then
    the one-hot encoded label is the float array: [0. 0. 1. 0.]

    Args:
        class_numbers: Array of integers with class-numbers.
            Assume the integers are from zero to num_classes-1 inclusive.

        num_classes: Number of classes. If None then use max(cls)-1.

    Return:
        2-dim array of shape: [len(cls), num_classes]
    """

    # Find the number of classes if None is provided.
    if num_classes is None:
        num_classes = np.max(class_numbers) - 1

    return np.eye(num_classes, dtype=float)[class_numbers]
